/*
 * This file is part of the AdvanceMAME project.
 *
 * Copyright (C) 1999-2002 Andrea Mazzoleni
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <zlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <assert.h>
#include <ctype.h>

/*************************************************************************************/
/* Error */

void error_io(void) {
	fprintf(stderr,"IO error\n");
}

void error_nomem(void) {
	fprintf(stderr,"Low memory\n");
}

void error_unsupported(const char* s) {
	fprintf(stderr,"%s\n",s);
}

void error_inval(const char* s) {
	fprintf(stderr,"%s\n",s);
}

void error_internal(int line) {
	fprintf(stderr,"Internal error at line %d\n",line);
}

int opt_verbose;
int opt_dx;
int opt_dy;
int opt_reduce;
int opt_expand;
int opt_level;

/*************************************************************************************/
/* MNG */

#define PNG_CN_IHDR 0x49484452
#define PNG_CN_PLTE 0x504C5445
#define PNG_CN_IDAT 0x49444154
#define PNG_CN_IEND 0x49454E44

#define MNG_CN_DHDR 0x44484452
#define MNG_CN_MHDR 0x4D484452
#define MNG_CN_MEND 0x4D454E44
#define MNG_CN_DEFI 0x44454649
#define MNG_CN_PPLT 0x50504c54
#define MNG_CN_MOVE 0x4d4f5645
#define MNG_CN_FRAM 0x4652414d

static unsigned char MNG_Signature[] = "\x8A\x4D\x4E\x47\x0D\x0A\x1A\x0A";

unsigned read_be_u32(unsigned char* v) {
	return v[0] << 24 | v[1] << 16 | v[2] << 8 | v[3];
}

void write_be_u32(unsigned char* v, unsigned i) {
	v[0] = (i >> 24) & 0xFF;
	v[1] = (i >> 16) & 0xFF;
	v[2] = (i >> 8) & 0xFF;
	v[3] = i & 0xFF;
}

void write_be_u16(unsigned char* v, unsigned i) {
	v[0] = (i >> 8) & 0xFF;
	v[1] = i & 0xFF;
}

int mng_write_chunk(FILE* f, unsigned* fc, unsigned chunk_type, unsigned char* chunk_data, unsigned chunk_length)
{
	unsigned char v[4];
	unsigned crc;

	write_be_u32(v, chunk_length);
	if (fwrite(v, 4, 1, f) != 1) {
		error_io();
		return -1;
	}
	*fc += 4;

	write_be_u32(v, chunk_type);
	if (fwrite(v, 4, 1, f) != 1) {
		error_io();
		return -1;
	}
	*fc += 4;

	crc = crc32(0, v, 4);
	if (chunk_length > 0) {
		if (fwrite(chunk_data, chunk_length, 1, f) != 1) {
			error_io();
			return -1;
		}
		*fc += chunk_length;

		crc = crc32(crc, chunk_data, chunk_length);
	}

	write_be_u32(v, crc);
	if (fwrite(v, 4, 1, f) != 1) {
		error_io();
		return -1;
	}
	*fc += 4;

	return 0;
}

int mng_read_chunk(FILE* f, unsigned char** data, unsigned* size, unsigned* type) {
	unsigned char cl[4];
	unsigned char ct[4];
	unsigned char cc[4];
	unsigned crc;

	if (fread(cl, 4, 1, f) != 1) {
		error_io();
		goto err;
	}

	*size = read_be_u32(cl);

	if (fread(ct, 4, 1, f) != 1) {
		error_io();
		goto err;
	}

	crc = crc32(0, ct, 4);

	*type = read_be_u32(ct);

	if (*size) {
		*data = malloc(*size);
		if (!*data) {
			error_nomem();
			goto err;
		}

		if (fread(*data, *size, 1, f) != 1) {
			error_io();
			goto err_data;
		}

		crc = crc32(crc, *data, *size);
	} else {
		*data = 0;
	}

	if (fread(cc, 4, 1, f) != 1) {
		error_io();
		goto err_data;
	}

        if (crc != read_be_u32(cc)) {
		error_inval("Invalid CRC");
		goto err_data;
	}

	return 0;
err_data:
	free(*data);
err:
	return -1;
}

int mng_read_signature(FILE* f) {
	unsigned char signature[8];

	if (fread(signature,8,1,f) != 1) {
		error_io();
		return -1;
	}

	if (memcmp(signature,MNG_Signature,8)!=0) {
		error_inval("Invalid signature");
		return -1;
	}

	return 0;
}

int mng_write_signature(FILE* f, unsigned* fc) {
	if (fwrite(MNG_Signature, 8, 1, f) != 1) {
		error_io();
		return -1;
	}
	*fc += 8;

	return 0;
}

/*************************************************************************************/
/* Read */

struct read {
	unsigned width; /* size of the frame */
	unsigned height;
	unsigned frequency; /* frame rate */

	unsigned char* pal_ptr; /* current palette */
	unsigned pal_size; /* size in bytes of the palette */

	unsigned char* img_ptr; /* pointer at the image */
	unsigned img_size; /* size in bytes of the image */
	unsigned img_scanline; /* size in bytes of the image scanline */
	unsigned img_pixel; /* size in bytes of the image pixel */
} READ;

int read_header(unsigned char* data, unsigned size) {
	unsigned simplicity;

	if (size != 28) {
		error_inval("Invalid size of the MHDR chunk");
		return -1;
	}

	simplicity = read_be_u32(data + 24);

	if (simplicity != 0x41 /* MNG-VLC without transparency */
		&& simplicity != 0x43 /* MNG-VLC without transparency + FRAME */) {
		error_unsupported("Only MNG-VLC files are supported");
		return -1;
	}

	READ.width = read_be_u32(data + 0);
	READ.height = read_be_u32(data + 4);
	READ.frequency = read_be_u32(data + 8);

	return 0;
}

int read_image(FILE* f, unsigned char* data, unsigned size) {
	unsigned fil_scanline;
	unsigned long fil_size;
	unsigned z_size;
	unsigned end_size;
	unsigned char* fil_ptr;
	unsigned char* z_ptr;
	unsigned char* end_ptr;
	unsigned type;
	unsigned i;

	if (size != 13) {
		error_inval("Invalid size of the IHDR chunk");
		return -1;
	}

	if (read_be_u32(data + 0) != READ.width) {
		error_unsupported("Only images with the exact frame size are supported");
		return -1;
	}

	if (read_be_u32(data + 4) != READ.height) {
		error_unsupported("Only images with the exact frame size are supported");
		return -1;
	}

	if (data[8] != 8) { /* bit depth */
		error_unsupported("Only the 8 bit depth is supported");
		return -1;
	}

	if (data[9] != 2 && data[9] != 3) { /* color type */
		error_unsupported("Only the color type 2 and 3 are supported");
		return -1;
	}

	if (data[9] == 2)
		READ.img_pixel = 3;
	else
		READ.img_pixel = 1;

	if (data[10] != 0) { /* compression */
		error_unsupported("Only the compression type 0 is supported");
		return -1;
	}

	if (data[11] != 0) { /* filter */
		error_unsupported("Only the filter type 0 is supported");
		return -1;
	}

	if (data[12] != 0) { /* interlace */
		error_unsupported("Only the not interlaced images are supported");
		return -1;
	}

	if (READ.img_pixel == 1) {
		if (mng_read_chunk(f, &READ.pal_ptr, &READ.pal_size, &type) != 0)
			return -1;

		if (type != PNG_CN_PLTE) {
			error_unsupported("The PLTE chunk is missing");
			return -1;
		}

		if (READ.pal_size % 3 != 0 || READ.pal_size / 3 > 256) {
			error_inval("Invalid palette size");
			return -1;
		}
	} else {
		READ.pal_ptr = 0;
		READ.pal_size = 0;
	}

	if (mng_read_chunk(f, &z_ptr, &z_size, &type) != 0) {
		return -1;
	}

	if (type != PNG_CN_IDAT) {
		error_unsupported("The IDAT chunk is missing");
		return -1;
	}

	if (mng_read_chunk(f, &end_ptr, &end_size, &type) != 0) {
		return -1;
	}

	if (type != PNG_CN_IEND) {
		error_unsupported("The IEND chunk is missing");
		return -1;
	}

	fil_scanline = READ.width * READ.img_pixel + 1;
	fil_size = READ.height * fil_scanline;
	fil_ptr = malloc(fil_size);

	if (uncompress(fil_ptr, &fil_size, z_ptr, z_size) != Z_OK) {
		error_inval("Corrupt compressed data");
		return -1;
	}

	READ.img_scanline = READ.width * READ.img_pixel;
	READ.img_size = READ.height * READ.img_scanline;
	READ.img_ptr = malloc(READ.img_size);

	for(i=0;i<READ.height;++i) {
		if (fil_ptr[i * fil_scanline] != 0) {
			error_unsupported("Only the line filter 0 (no filter) is supported");
			return -1;
		}
		memcpy( &READ.img_ptr[i * READ.img_scanline], &fil_ptr[i * fil_scanline + 1], READ.img_scanline );
	}

	free(end_ptr);
	free(z_ptr);
	free(fil_ptr);

	return 0;
}

/*************************************************************************************/
/* Compare */

unsigned compare_single1(unsigned width, unsigned height, unsigned char* p0, unsigned char* p1, unsigned line) {
	unsigned i,j;
	unsigned count = 0;

	for(i=0;i<height;++i) {
#if 0
		for(j=0;j<width;++j) {
			if (p0[0] == p1[0])
				++count;
			++p0;
			++p1;
		}
#else
		assert( sizeof(unsigned) == 4 );
		j = width;
		while (j >= 4) {
			unsigned v0 = *(unsigned*)p0;
			unsigned v1 = *(unsigned*)p1;
			v0 ^= v1;
			if (v0) {
				if (v0 & 0xFF)
					++count;
				if (v0 & 0xFF00)
					++count;
				if (v0 & 0xFF0000)
					++count;
				if (v0 & 0xFF000000)
					++count;
			} else {
				count += 4;
			}
			p0 += 4;
			p1 += 4;
			j -= 4;
		}
		while (j > 0) {
			if (p0[0] == p1[0])
				++count;
			++p0;
			++p1;
			--j;
		}
#endif
		p0 += line - width;
		p1 += line - width;
	}

	return count;
}

unsigned compare_single3(unsigned width, unsigned height, unsigned char* p0, unsigned char* p1, unsigned line) {
	unsigned i,j;
	unsigned count = 0;

	for(i=0;i<height;++i) {
		for(j=0;j<width;++j) {
			if (p0[0] == p1[0] && p0[1] == p1[1] && p0[2] == p1[2])
				++count;
			p0 += 3;
			p1 += 3;
		}
		p0 += line - width * 3;
		p1 += line - width * 3;
	}

	return count;
}

unsigned compare_shift(int x, int y, unsigned width, unsigned height, unsigned char* p0, unsigned char* p1, unsigned pixel, unsigned line) {
	unsigned count;
	int dx = width - abs(x);
	int dy = height - abs(y);

	if (x < 0)
		p1 += -x * pixel;
	else
		p0 += x * pixel;

	if (y < 0)
		p1 += -y * line;
	else
		p0 += y * line;

	if (pixel == 1)
		count = compare_single1(dx, dy, p0, p1, line);
	else
		count = compare_single3(dx, dy, p0, p1, line);

	return count;
}

void compare(int* x, int* y, unsigned width, unsigned height, unsigned char* p0, unsigned char* p1, unsigned pixel, unsigned line) {
	int i,j;
	int best_x;
	int best_y;
	unsigned best_count;
	int prev_x;
	int prev_y;
	unsigned prev_count;
	unsigned center_count;
	unsigned total;

	best_x = 0;
	best_y = 0;
	best_count = compare_shift(0, 0, width, height, p0, p1, pixel, line);

	prev_count = 0;
	prev_x = 0;
	prev_y = 0;

	center_count = best_count;

	total = width * height;

	for(i=-opt_dx;i<=opt_dx;++i) {
	for(j=-opt_dy;j<=opt_dy;++j) {
	if (j || i) {
		unsigned count;

		count = compare_shift(i, j, width, height, p0, p1, pixel, line);
		if (count > best_count) {
			prev_count = best_count;
			prev_x = best_x;
			prev_y = best_y;
			best_count = count;
			best_x = i;
			best_y = j;
		} else if (count > prev_count) {
			prev_count = count;
			prev_x = i;
			prev_y = j;
		}
	}
	}
	}

	/* if too small difference than fixed image don't move */
	if (best_count < total / 3) {

		if (opt_verbose)
			printf("disable, best %3d %3d %6d, 2'nd %3d %3d %6d, fix %6d\n", best_x, best_y, best_count * 100 / total, prev_x, prev_y, prev_count * 100 / total, center_count * 100 / total);

		*x = 0;
		*y = 0;
	} else {

		if (opt_verbose)
			printf("best %3d %3d %6d, 2'nd %3d %3d %6d, fix %6d\n", best_x, best_y, best_count * 100 / total, prev_x, prev_y, prev_count * 100 / total, center_count * 100 / total);

		*x = best_x;
		*y = best_y;
	}
}

/*************************************************************************************/
/* Analyze */

struct scroll_coord {
	int x;
	int y;
};

struct scroll {
	struct scroll_coord* map; /* vector of scrolling shift */
	unsigned mac; /* space used */
	unsigned max; /* space allocated */
	int x; /* start position of the first image in the scroll buffer */
	int y;
	unsigned width; /* additional size of the scroll buffer */
	unsigned height;
};

struct analyze {
	unsigned char* pre_ptr; /* previous image */
	unsigned char* cur_ptr; /* current image */
} ANALYZE;

int analyze_insert(struct scroll* sc, int x, int y) {
	if (sc->mac == sc->max) {
		sc->max *= 2;
		if (!sc->max)
			sc->max = 64;
		sc->map = realloc(sc->map, sc->max * sizeof(struct scroll_coord));
		if (!sc->map) {
			error_nomem();
			return -1;
		}
	}

	sc->map[sc->mac].x = x;
	sc->map[sc->mac].y = y;
	++sc->mac;

	return 0;
}

int analyze_image(FILE* f, struct scroll* sc, unsigned char* data, unsigned size) {
	if (read_image(f, data,size) != 0)
		return -1;

	free(ANALYZE.pre_ptr);
	ANALYZE.pre_ptr = ANALYZE.cur_ptr;
	ANALYZE.cur_ptr = READ.img_ptr;

	if (ANALYZE.pre_ptr && ANALYZE.cur_ptr) {
		int x;
		int y;
		compare(&x, &y, READ.width, READ.height, ANALYZE.pre_ptr, ANALYZE.cur_ptr, READ.img_pixel, READ.img_scanline);
		if (analyze_insert(sc, x,y) != 0)
			return -1;
	} else {
		if (analyze_insert(sc, 0,0) != 0)
			return -1;
	}

	free(READ.pal_ptr);

	return 0;
}

void analyze_range(struct scroll* sc) {
	int px = 0;
	int py = 0;
	int min_x = 0;
	int min_y = 0;
	int max_x = 0;
	int max_y = 0;
	unsigned i;

	for(i=0;i<sc->mac;++i) {
		px += sc->map[i].x;
		py += sc->map[i].y;
		if (px < min_x)
			min_x = px;
		if (py < min_y)
			min_y = py;
		if (px > max_x)
			max_x = px;
		if (py > max_y)
			max_y = py;
	}

	sc->x = -min_x;
	sc->y = -min_y;
	sc->width = max_x - min_x;
	sc->height = max_y - min_y;
}

int analyze_f(FILE* f, struct scroll* sc) {
	unsigned type;
	unsigned size;
	unsigned char* data;
	unsigned counter;

	if (mng_read_signature(f) != 0) {
		return -1;
	}

	ANALYZE.pre_ptr = 0;
	ANALYZE.cur_ptr = 0;

	sc->map = 0;
	sc->mac = 0;
	sc->max = 0;

	counter = 0;

	do {
		if (mng_read_chunk(f, &data, &size, &type) != 0)
			return -1;

		switch (type) {
			case MNG_CN_MHDR :
				if (read_header(data, size) != 0)
					return -1;
				break;
			case PNG_CN_IHDR :
				if (analyze_image(f, sc, data, size) != 0)
					return -1;
				++counter;
				fprintf(stdout,"frame %d   \r",counter);
				fflush(stdout);
				break;
			case MNG_CN_MEND :
				break;
			case MNG_CN_FRAM :
				break;
			default:
				error_unsupported("Unknow chunk detected");
				return -1;
		}

		free(data);

	} while (type != MNG_CN_MEND);

	free(ANALYZE.pre_ptr);
	free(ANALYZE.cur_ptr);

	analyze_range(sc);

	fprintf(stdout,"frame %d    \n",counter);

	if (sc->width || sc->height)
		printf("pos %dx%d scroll %dx%d\n",sc->x,sc->y,sc->width,sc->height);
	else
		printf("no scroll\n");

	return 0;
}

int analyze_file(const char* file, struct scroll* sc) {
	FILE* f;

	f = fopen(file,"rb");
	if (!f) {
		error_io();
		return -1;
	}

	if (analyze_f(f,sc) != 0) {
		fclose(f);
		return -1;
	}

	fclose(f);

	return 0;
}

/*************************************************************************************/
/* Write */

struct write {
	int first; /* first image flag */

	unsigned width; /* frame size in pixel */
	unsigned height;
	unsigned pixel; /* pixel size in bytes */

	int scroll_x; /* initial position of the first image in the scroll buffer */
	int scroll_y;
	unsigned scroll_width; /* extra scroll buffer size */
	unsigned scroll_height;

	unsigned line; /* line size in bytes */
	unsigned char* scroll_ptr; /* global image data */

	unsigned char* current_ptr; /* pointer at the current frame in the scroll buffer */
	int current_x; /* current scroll position */
	int current_y;

	unsigned char pal_ptr[256*3]; /* palette */
	unsigned pal_size; /* palette size in bytes */
	int pal_used[256]; /* flag vector for the used entries in the palette */
} WRITE;

int write_reduce(unsigned char** out_ptr, unsigned* out_scanline, unsigned char* ovr_ptr, unsigned* ovr_used, unsigned char* img_ptr, unsigned img_scanline) {
	unsigned char col_ptr[256*3];
	unsigned col_mapped[256];
	unsigned col_count;
	unsigned i,j,k;
	unsigned char* new_ptr;
	unsigned new_scanline;

	/* build the new palette */
	col_count = 0;
	for(i=0;i<WRITE.height;++i) {
		unsigned char* p0 = img_ptr + i * img_scanline;
		for(j=0;j<WRITE.width;++j) {
			for(k=0;k<col_count;++k) {
				if (col_ptr[k*3] == p0[0] && col_ptr[k*3+1] == p0[1] && col_ptr[k*3+2] == p0[2])
					break;
			}
			if (k == col_count) {
				if (col_count == 256)
					return -1; /* too many colors */
				col_ptr[col_count*3] = p0[0];
				col_ptr[col_count*3+1] = p0[1];
				col_ptr[col_count*3+2] = p0[2];
				++col_count;
			}
			p0 += 3;
		}
	}

	for(i=0;i<256;++i) {
		ovr_used[i] = 0;
		col_mapped[i] = 0;
	}
	memcpy(ovr_ptr, WRITE.pal_ptr, 256*3);

	/* map colors already present in the old palette */
	for(i=0;i<col_count;++i) {
		for(k=0;k<256;++k) {
			if (!ovr_used[k] && ovr_ptr[k*3]==col_ptr[i*3] && ovr_ptr[k*3+1]==col_ptr[i*3+1] && ovr_ptr[k*3+2]==col_ptr[i*3+2])
				break;
		}
		if (k<256) {
			ovr_used[k] = 1;
			col_mapped[i] = 1;
		}
	}

	/* map colors not present in the old palette */
	for(i=0;i<col_count;++i) {
		if (!col_mapped[i]) {
			/* search the first free space */
			for(k=0;k<256;++k)
				if (!ovr_used[k])
					break;
			ovr_used[k] = 1;
			ovr_ptr[k*3] = col_ptr[i*3];
			ovr_ptr[k*3+1] = col_ptr[i*3+1];
			ovr_ptr[k*3+2] = col_ptr[i*3+2];
		}
	}

	/* create the new bitmap */
	new_scanline = WRITE.width;
	new_ptr = malloc(WRITE.height * new_scanline);
	for(i=0;i<WRITE.height;++i) {
		unsigned char* p0 = new_ptr + i*new_scanline;
		unsigned char* p1 = img_ptr + i*img_scanline;
		for(j=0;j<WRITE.width;++j) {
			for(k=0;;++k)
				if (ovr_ptr[k*3]==p1[0] && ovr_ptr[k*3+1]==p1[1] && ovr_ptr[k*3+2]==p1[2])
					break;
			*p0 = k;
			++p0;
			p1 += 3;
		}
	}

	*out_ptr = new_ptr;
	*out_scanline = new_scanline;

	return 0;
}

void write_expand(unsigned char** out_ptr, unsigned* out_scanline, unsigned char* img_ptr, unsigned img_scanline, unsigned char* pal_ptr) {
	unsigned char* new_ptr;
	unsigned new_scanline;
	unsigned i,j;

	/* create the new bitmap */
	new_scanline = WRITE.width * 3;
	new_ptr = malloc(WRITE.height * new_scanline);

	for(i=0;i<WRITE.height;++i) {
		unsigned char* p0 = new_ptr + i*new_scanline;
		unsigned char* p1 = img_ptr + i*img_scanline;
		for(j=0;j<WRITE.width;++j) {
			unsigned k = 3 * *p1;
			p0[0] = pal_ptr[k];
			p0[1] = pal_ptr[k+1];
			p0[2] = pal_ptr[k+2];
			p0 += 3;
			++p1;
		}
	}

	*out_ptr = new_ptr;
	*out_scanline = new_scanline;
}

int write_header(FILE* f, unsigned* fc, unsigned width, unsigned height, unsigned frequency, int scroll_x, int scroll_y, unsigned scroll_width, unsigned scroll_height) {
	unsigned simplicity;
	unsigned char mhdr[28];

	if (mng_write_signature(f,fc) != 0) {
		return -1;
	}

	simplicity = (1 << 0) /* Enable flags */
		| (1 << 1) /* Basic features */
		| (1 << 2) /* Extended features */
		| (1 << 5) /* Delta PNG */
		| (1 << 6) /* Enable flags */
		| (1 << 9); /* Object buffers must be stored */

	write_be_u32(mhdr + 0, width); /* width */
	write_be_u32(mhdr + 4, height); /* height */
	write_be_u32(mhdr + 8, frequency ); /* ticks per second */
	write_be_u32(mhdr + 12, 0 ); /* nominal layer count */
	write_be_u32(mhdr + 16, 0 ); /* nominal frame count */
	write_be_u32(mhdr + 20, 0 ); /* nominal play time */
	write_be_u32(mhdr + 24, simplicity); /* simplicity profile */

	if (mng_write_chunk(f, fc, MNG_CN_MHDR, mhdr, sizeof(mhdr))!=0)
		return -1;

	WRITE.width = width;
	WRITE.height = height;
	WRITE.pixel = 0;

	WRITE.scroll_x = scroll_x;
	WRITE.scroll_y = scroll_y;
	WRITE.scroll_width = scroll_width;
	WRITE.scroll_height = scroll_height;
	WRITE.scroll_ptr = 0;

	WRITE.first = 1;

	return 0;
}

void compute_image(unsigned char** out_ptr, unsigned* out_size, unsigned char* img_ptr, unsigned img_scanline, unsigned x, unsigned y, unsigned dx, unsigned dy) {
	unsigned char* fil_ptr;
	unsigned fil_size;
	unsigned fil_scanline;
	unsigned char* z_ptr;
	unsigned long z_size;
	unsigned i;
	unsigned char* p0;

	fil_scanline = dx * WRITE.pixel + 1;
	fil_size = dy * fil_scanline;
	z_size = fil_size * 11 / 10 + 12;

	fil_ptr = malloc(fil_size);
	z_ptr = malloc(z_size);

	p0 = fil_ptr;

	for(i=0;i<dy;++i) {
		unsigned char* p1 = &img_ptr[x * WRITE.pixel + (i+y) * img_scanline];
		*p0++ = 0;
		memcpy(p0, p1, dx * WRITE.pixel);
		p0 += dx * WRITE.pixel;
	}

	assert(p0 == fil_ptr + fil_size);

	if (compress2(z_ptr, &z_size, fil_ptr, fil_size, opt_level) != Z_OK) {
		abort();
	}

	free(fil_ptr);

	*out_ptr = z_ptr;
	*out_size = z_size;
}

void compute_image_delta(unsigned char** out_ptr, unsigned* out_size, unsigned char* img_ptr, unsigned img_scanline, unsigned x, unsigned y, unsigned dx, unsigned dy) {
	unsigned char* fil_ptr;
	unsigned fil_size;
	unsigned fil_scanline;
	unsigned char* z_ptr;
	unsigned long z_size;
	unsigned i;
	unsigned char* p0;

	fil_scanline = dx * WRITE.pixel + 1;
	fil_size = dy * fil_scanline;
	z_size = fil_size * 11 / 10 + 12;

	fil_ptr = malloc(fil_size);
	z_ptr = malloc(z_size);

	p0 = fil_ptr;

	for(i=0;i<dy;++i) {
		unsigned j;
		unsigned char* p1 = &img_ptr[x * WRITE.pixel + (i+y) * img_scanline];
		unsigned char* p2 = &WRITE.current_ptr[x * WRITE.pixel + (i+y) * WRITE.line];

		*p0++ = 0;
		for(j=0;j<dx*WRITE.pixel;++j)
			*p0++ = *p1++ - *p2++;
	}

	assert(p0 == fil_ptr + fil_size);

	if (compress2(z_ptr, &z_size, fil_ptr, fil_size, opt_level) != Z_OK) {
		abort();
	}

	free(fil_ptr);

	*out_ptr = z_ptr;
	*out_size = z_size;
}

int row_equal(unsigned y, unsigned char* img_ptr, unsigned img_scanline) {
	unsigned char* p0;
	unsigned char* p1;

	p0 = WRITE.current_ptr + y * WRITE.line;
	p1 = img_ptr + y * img_scanline;

	return memcmp(p0, p1, WRITE.width * WRITE.pixel) == 0;
}

int col_equal(unsigned x, unsigned char* img_ptr, unsigned img_scanline) {
	unsigned char* p0;
	unsigned char* p1;
	unsigned i;

	p0 = WRITE.current_ptr + x * WRITE.pixel;
	p1 = img_ptr + x * WRITE.pixel;

	if (WRITE.pixel == 1) {
		for(i=0;i<WRITE.height;++i) {
			if (p0[0] != p1[0])
				return 0;
			p0 += WRITE.line;
			p1 += img_scanline;
		}
	} else {
		for(i=0;i<WRITE.height;++i) {
			if (p0[0] != p1[0] || p0[1] != p1[1] || p0[2] != p1[2])
				return 0;
			p0 += WRITE.line;
			p1 += img_scanline;
		}
	}

	return 1;
}

void compute_image_range(unsigned* out_x, unsigned* out_y, unsigned* out_dx, unsigned* out_dy, unsigned char* img_ptr, unsigned img_scanline) {
	unsigned x;
	unsigned dx;
	unsigned y;
	unsigned dy;

	x = 0;
	while (x < WRITE.width && col_equal(x, img_ptr, img_scanline))
		++x;

	dx = WRITE.width - x;
	while (dx > 0 && col_equal(x + dx - 1, img_ptr, img_scanline))
		--dx;

	y = 0;
	while (y < WRITE.height && row_equal(y, img_ptr, img_scanline))
		++y;

	dy = WRITE.height - y;
	while (dy > 0 && row_equal(y + dy - 1, img_ptr, img_scanline))
		--dy;

	*out_x = x;
	*out_y = y;
	*out_dx = dx;
	*out_dy = dy;
}

void compute_palette_delta(unsigned char** out_ptr, unsigned* out_size, unsigned char* pal_ptr, unsigned pal_size) {
	unsigned i;
	unsigned char* dst_ptr;
	unsigned dst_size;

	dst_ptr = malloc(pal_size * 2);
	dst_size = 0;

	dst_ptr[dst_size++] = 0; /* replacement */

	i = 0;
	while (i<pal_size) {
		unsigned j;

		while (i < pal_size && WRITE.pal_ptr[i] == pal_ptr[i] && WRITE.pal_ptr[i+1] == pal_ptr[i+1] && WRITE.pal_ptr[i+2] == pal_ptr[i+2])
			i += 3;

		if (i == pal_size)
			break;

		j = i;

		while (j < pal_size && (WRITE.pal_ptr[j] != pal_ptr[j] || WRITE.pal_ptr[j+1] != pal_ptr[j+1] || WRITE.pal_ptr[j+2] != pal_ptr[j+2]))
			j += 3;

		dst_ptr[dst_size++] = i / 3; /* first index */
		dst_ptr[dst_size++] = (j / 3) - 1; /* last index */

		while (i < j) {
			dst_ptr[dst_size++] = pal_ptr[i++];
			dst_ptr[dst_size++] = pal_ptr[i++];
			dst_ptr[dst_size++] = pal_ptr[i++];
		}
	}

	if (dst_size == 1) {
		*out_ptr = 0;
		*out_size = 0;
		free(dst_ptr);
	} else {
		*out_ptr = dst_ptr;
		*out_size = dst_size;
	}
}

void write_store(unsigned char* img_ptr, unsigned img_scanline, unsigned char* pal_ptr, unsigned pal_size) {
	unsigned i;

	if (pal_size) {
		memcpy(WRITE.pal_ptr, pal_ptr, pal_size);
		memset(WRITE.pal_ptr + pal_size, 0, 256*3 - pal_size);
		if (pal_size > WRITE.pal_size)
			WRITE.pal_size = pal_size;
	}

	for(i=0;i<WRITE.height;++i) {
		memcpy( &WRITE.current_ptr[i * WRITE.line], &img_ptr[i * img_scanline], WRITE.width * WRITE.pixel);
	}
}

int write_first(FILE* f, unsigned* fc) {
	unsigned char defi[12];
	unsigned defi_size;
	unsigned char ihdr[13];
	unsigned char* z_ptr;
	unsigned z_size;

	write_be_u16(defi + 0, 1); /* object id */
	defi[2] = 0; /* visible */
	defi[3] = 1; /* concrete */
	if (WRITE.scroll_x!=0 || WRITE.scroll_y!=0) {
		write_be_u32(defi + 4, - WRITE.scroll_x);
		write_be_u32(defi + 8, - WRITE.scroll_y);
		defi_size = 12;
	} else
		defi_size = 4;
	if (mng_write_chunk(f, fc, MNG_CN_DEFI, defi, defi_size)!=0)
		return -1;

	write_be_u32(ihdr + 0, WRITE.width + WRITE.scroll_width);
	write_be_u32(ihdr + 4, WRITE.height + WRITE.scroll_height);
	ihdr[8] = 8; /* bit depth */
	if (WRITE.pixel == 1)
		ihdr[9] = 3; /* color type */
	else
		ihdr[9] = 2; /* color type */
	ihdr[10] = 0; /* compression */
	ihdr[11] = 0; /* filter */
	ihdr[12] = 0; /* interlace */

	if (mng_write_chunk(f, fc, PNG_CN_IHDR, ihdr, sizeof(ihdr))!=0) {
		return -1;
	}

	if (WRITE.pal_size) {
		if (mng_write_chunk(f, fc, PNG_CN_PLTE, WRITE.pal_ptr, WRITE.pal_size)!=0) {
			return -1;
		}
	}

	compute_image(&z_ptr, &z_size, WRITE.scroll_ptr, WRITE.line, 0, 0, WRITE.width + WRITE.scroll_width, WRITE.height + WRITE.scroll_height);

	if (mng_write_chunk(f, fc, PNG_CN_IDAT, z_ptr, z_size)!=0) {
		return -1;
	}

	free(z_ptr);

	if (mng_write_chunk(f, fc, PNG_CN_IEND, 0, 0)!=0)
		return -1;

	return 0;
}

int write_move(FILE* f, unsigned* fc, int shift_x, int shift_y) {
	unsigned char move[13];

	if (shift_x!=0 || shift_y!=0) {

		write_be_u16(move + 0, 1); /* id 1 */
		write_be_u16(move + 2, 1); /* id 1 */
		move[4] = 1; /* adding */
		write_be_u32(move + 5, - shift_x);
		write_be_u32(move + 9, - shift_y);

		if (mng_write_chunk(f, fc, MNG_CN_MOVE, move, sizeof(move))!=0)
			return -1;

	}

	return 0;
}

int write_delta_image(FILE* f, unsigned* fc, unsigned char* img_ptr, unsigned img_scanline, unsigned char* pal_ptr, unsigned pal_size) {
	unsigned x,y,dx,dy;
	unsigned char* pal_d_ptr;
	unsigned pal_d_size;
	unsigned char* pal_r_ptr;
	unsigned pal_r_size;
	unsigned char* z_d_ptr;
	unsigned z_d_size;
	unsigned char* z_r_ptr;
	unsigned z_r_size;

	if (pal_ptr && pal_size) {
		compute_palette_delta(&pal_d_ptr, &pal_d_size, pal_ptr, pal_size);

		pal_r_ptr = malloc(pal_size);
		memcpy(pal_r_ptr, pal_ptr, pal_size);
                pal_r_size = pal_size;
	} else {
		pal_d_ptr = 0;
		pal_d_size = 0;
		pal_r_ptr = 0;
		pal_r_size = 0;
	}

	compute_image_range(&x, &y, &dx, &dy, img_ptr, img_scanline);

	if (dx && dy) {
		compute_image_delta(&z_d_ptr, &z_d_size, img_ptr, img_scanline, x, y, dx, dy);
		compute_image(&z_r_ptr, &z_r_size, img_ptr, img_scanline, x, y, dx, dy);
	} else {
		z_d_ptr = 0;
		z_d_size = 0;
		z_r_ptr = 0;
		z_r_size = 0;
	}

	if (1) {
		char dhdr[20];
		unsigned dhdr_size;

		write_be_u16(dhdr + 0, 1); /* object id */
		dhdr[2] = 1; /* png image */
		if (z_d_size) {
			if (z_d_size < z_r_size) {
				dhdr[3] = 1; /* block pixel addition */
				dhdr_size = 20;
			} else {
				if (dx == WRITE.width && dy == WRITE.height && WRITE.scroll_width == 0 && WRITE.scroll_height == 0) {
					dhdr[3] = 0; /* entire image replacement */
					dhdr_size = 12;
				} else {
					dhdr[3] = 4; /* block pixel replacement */
					dhdr_size = 20;
				}
			}
		} else {
			dhdr[3] = 7; /* no change to pixel data */
			dhdr_size = 4;
		}

		write_be_u32(dhdr + 4, dx);
		write_be_u32(dhdr + 8, dy);
		write_be_u32(dhdr + 12, x + WRITE.current_x);
		write_be_u32(dhdr + 16, y + WRITE.current_y);

		if (mng_write_chunk(f, fc, MNG_CN_DHDR, dhdr, dhdr_size)!=0)
			return -1;

		if (pal_d_size) {
			if (pal_d_size < pal_r_size) {
				if (mng_write_chunk(f, fc, MNG_CN_PPLT, pal_d_ptr, pal_d_size)!=0)
					return -1;
			} else {
				if (mng_write_chunk(f, fc, PNG_CN_PLTE, pal_r_ptr, pal_r_size)!=0)
					return -1;
			}
		}

		if (z_d_size) {
			if (z_d_size < z_r_size) {
				if (mng_write_chunk(f, fc, PNG_CN_IDAT, z_d_ptr, z_d_size)!=0)
					return -1;
			} else {
				if (mng_write_chunk(f, fc, PNG_CN_IDAT, z_r_ptr, z_r_size)!=0)
					return -1;
			}
		}

		if (mng_write_chunk(f, fc, PNG_CN_IEND, 0, 0)!=0)
			return -1;
	}

	free(z_r_ptr);
	free(z_d_ptr);
	free(pal_r_ptr);
	free(pal_d_ptr);

	write_store(img_ptr, img_scanline, pal_ptr, pal_size);

	return 0;
}

int write_image_raw(FILE* f, unsigned* fc, unsigned width, unsigned height, unsigned pixel, unsigned char* img_ptr, unsigned img_scanline, unsigned char* pal_ptr, unsigned pal_size, int shift_x, int shift_y) {
	if (width != WRITE.width) {
		error_internal(__LINE__);
		return -1;
	}

	if (height != WRITE.height) {
		error_internal(__LINE__);
		return -1;
	}

	if (WRITE.first) {
		if (shift_x!=0 || shift_y!=0) {
			error_internal(__LINE__);
			return -1;
		}

		WRITE.first = 0;
		WRITE.pixel = pixel;
		WRITE.line = pixel * (WRITE.width + WRITE.scroll_width);
		WRITE.scroll_ptr = malloc( (WRITE.height + WRITE.scroll_height) * WRITE.line);
		WRITE.current_ptr = WRITE.scroll_ptr + WRITE.scroll_x * WRITE.pixel + WRITE.scroll_y * WRITE.line;
		WRITE.current_x = WRITE.scroll_x;
		WRITE.current_y = WRITE.scroll_y;

		memset(WRITE.scroll_ptr, 0, (WRITE.height + WRITE.scroll_height) * WRITE.line);
		memset(WRITE.pal_ptr,0, 256*3);

		write_store(img_ptr, img_scanline, pal_ptr, pal_size);

		return write_first(f, fc);
	} else {
		WRITE.current_ptr += shift_x * WRITE.pixel + shift_y * WRITE.line;
		WRITE.current_x += shift_x;
		WRITE.current_y += shift_y;

		if (write_move(f, fc, shift_x, shift_y) != 0)
			return -1;

		return write_delta_image(f, fc, img_ptr, img_scanline, pal_ptr, pal_size);
	}
}

int write_image(FILE* f, unsigned* fc, unsigned width, unsigned height, unsigned pixel, unsigned char* img_ptr, unsigned img_scanline, unsigned char* pal_ptr, unsigned pal_size, int shift_x, int shift_y) {
	if (pixel == 1) {
		unsigned char* new_ptr;
		unsigned new_scanline;
		int r;

		if (!opt_expand) {
			return write_image_raw(f, fc, width, height, pixel, img_ptr, img_scanline, pal_ptr, pal_size, shift_x, shift_y);
		}

		write_expand(&new_ptr, &new_scanline, img_ptr, img_scanline, pal_ptr);

		r = write_image_raw(f, fc, width, height, 3, new_ptr, new_scanline, 0, 0, shift_x, shift_y);

		free(new_ptr);

		return r;
	} else if (pixel == 3) {
		unsigned char ovr_ptr[256*3];
		unsigned ovr_used[256];
		unsigned char* new_ptr;
		unsigned new_scanline;
		int r;

		if (!opt_reduce) {
			return write_image_raw(f, fc, width, height, pixel, img_ptr, img_scanline, 0, 0, shift_x, shift_y);
		}

		if (WRITE.first || WRITE.pixel == 3) {
			memset(WRITE.pal_used, 0, sizeof(WRITE.pal_used));
			memset(WRITE.pal_ptr, 0, sizeof(WRITE.pal_ptr));
		}

		if (write_reduce(&new_ptr, &new_scanline, ovr_ptr, ovr_used, img_ptr, img_scanline) != 0) {
			/* reduction failed */
			if (WRITE.first) {
				return write_image_raw(f, fc, width, height, pixel, img_ptr, img_scanline, 0, 0, shift_x, shift_y);
			} else {
				fprintf(stderr,"Color reduction failed not at the first frame. Don't use the -r switch.\n");
				return -1;
			}
		}

		r = write_image_raw(f, fc, width, height, 1, new_ptr, new_scanline, ovr_ptr, 256*3, shift_x, shift_y);

		memcpy(WRITE.pal_used, ovr_used, sizeof(WRITE.pal_used));

		free(new_ptr);

		return r;
	}

	return -1;
}

int write_footer(FILE* f, unsigned* fc) {
	free(WRITE.scroll_ptr);

	if (mng_write_chunk(f, fc, MNG_CN_MEND, 0, 0)!=0)
		return -1;

	return 0;
}

/*************************************************************************************/
/* Conversion */

int convert_frame(FILE* f, unsigned* fc, unsigned char* data, unsigned size) {
	if (mng_write_chunk(f, fc, MNG_CN_FRAM, data, size) != 0)
		return -1;
	return 0;
}

int convert_header(FILE* f, unsigned* fc, unsigned char* data, unsigned size, struct scroll* sc) {
	if (read_header(data,size) != 0)
		return -1;

	if (sc) {
		if (write_header(f, fc, READ.width, READ.height, READ.frequency, sc->x, sc->y, sc->width, sc->height) != 0) {
			return -1;
		}
	} else {
		if (write_header(f, fc, READ.width, READ.height, READ.frequency, 0, 0, 0, 0) != 0) {
			return -1;
		}
	}

	return 0;
}

int convert_image(FILE* f_out, unsigned* fc, FILE* f_in, unsigned char* data, unsigned size, struct scroll_coord* scc) {

	if (read_image(f_in, data, size) != 0) {
		return -1;
	}

	if (scc) {
		if (write_image(f_out, fc, READ.width, READ.height, READ.img_pixel, READ.img_ptr, READ.img_scanline, READ.pal_ptr, READ.pal_size, scc->x, scc->y) != 0) {
			return -1;
		}
	} else {
		if (write_image(f_out, fc, READ.width, READ.height, READ.img_pixel, READ.img_ptr, READ.img_scanline, READ.pal_ptr, READ.pal_size, 0, 0) != 0) {
			return -1;
		}
	}

	free(READ.pal_ptr);
	free(READ.img_ptr);

	return 0;
}

int convert_f(FILE* f_in, FILE* f_out, unsigned* filec, unsigned* framec, struct scroll* sc) {
	unsigned type;
	unsigned size;
	unsigned char* data;
	unsigned counter;

	if (mng_read_signature(f_in) != 0) {
		return -1;
	}

	*filec = 0;
	counter = 0;

	do {
		if (mng_read_chunk(f_in, &data, &size, &type) != 0)
			return -1;

		switch (type) {
			case MNG_CN_MHDR :
				if (convert_header(f_out, filec, data, size, sc) != 0)
					return -1;
				break;
			case PNG_CN_IHDR :
				if (sc) {
					if (counter >= sc->mac) {
						error_internal(__LINE__);
						return -1;
					}
					if (convert_image(f_out, filec, f_in, data, size, &sc->map[counter]) != 0)
						return -1;
				} else {
					if (convert_image(f_out, filec, f_in, data, size, 0) != 0)
						return -1;
				}
				++counter;
				fprintf(stdout,"frame %d, size %d    \r", counter, *filec / counter);
				fflush(stdout);
				break;
			case MNG_CN_MEND :
				break;
			case MNG_CN_FRAM :
				if (convert_frame(f_out, filec, data, size) != 0)
					return -1;
				break;
				break;
			default:
				error_unsupported("Unsupported chunck detected");
				return -1;
		}

		free(data);

	} while (type != MNG_CN_MEND);

	if (write_footer(f_out, filec) != 0)
		return -1;

	fprintf(stdout,"frame %d, size %d    \n", counter, *filec / counter);

	*framec = counter;

	return 0;
}

int convert_file(const char* file_from, const char* file_to, struct scroll* sc) {
	FILE* f_in;
	FILE* f_out;
	unsigned filec;
	unsigned framec;

	f_in = fopen(file_from,"rb");
	if (!f_in) {
		error_io();
		return -1;
	}

	f_out = fopen(file_to,"wb");
	if (!f_in) {
		error_io();
		fclose(f_in);
		return -1;
	}

	if (convert_f(f_in,f_out, &filec, &framec, sc) != 0) {
		fclose(f_in);
		fclose(f_out);
		return -1;
	}

	fclose(f_in);
	fclose(f_out);

	return 0;
}

/***************************************************************************/
/* Main */

void error(void) {
	fprintf(stderr,"zmng v1.1 Copyright 2002 Andrea Mazzoleni\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"Syntax: zmng [-N] [-s NxM] [-r] FROM_FILE TO_FILE\n");
	fprintf(stderr,"\nOptions:\n");
	fprintf(stderr,"\t-N\tSelect the compression level.\n");
	fprintf(stderr,"\t\t-0 no, -1 low, -6 default, -9 high.\n");
	fprintf(stderr,"\t-s NxM\tEnable the scroll optimization with a pattern search.\n");
	fprintf(stderr,"\t\tfrom -Nx-M to NxM. Example: -s 4x6\n");
	fprintf(stderr,"\t-r\tConversion from 24 bit to palettized 8 bit.\n");
	fprintf(stderr,"\t-e\tConversion from 8 bit to 24 bit.\n");
	exit(EXIT_FAILURE);
}

int main(int argc, char* argv[]) {
	int i;
	struct scroll sc;
	int opt_scroll = 0;
	const char* opt_src = 0;
	const char* opt_dst = 0;
	opt_reduce = 0;
	opt_level = 6;
	opt_expand = 0;
	opt_verbose = 0;
	opt_dx = 0;
	opt_dy = 0;

	for(i=1;i<argc;++i) {
		if (strcmp(argv[i],"-s")==0 && i+1<argc) {
			int n;
			opt_dx = 0;
			opt_dy = 0;
			n = sscanf(argv[i+1],"%dx%d",&opt_dx,&opt_dy);
			if (n != 2)
				error();
			if (opt_dx < 0 || opt_dy < 0)
				error();
			if (opt_dx == 0 && opt_dy == 0)
				error();
			++i;
			opt_scroll = 1;
		} else if (argv[i][0] == '-' && isdigit(argv[i][1]) && argv[i][2]==0) {
			opt_level = argv[i][1] - '0';
		} else if (strcmp(argv[i],"-r")==0) {
			opt_reduce = 1;
			opt_expand = 0;
		} else if (strcmp(argv[i],"-e")==0) {
			opt_reduce = 0;
			opt_expand = 1;
		} else if (strcmp(argv[i],"-v")==0) {
			opt_verbose = 1;
		} else if (argv[i][0]=='-') {
			error();
		} else if (!opt_src) {
			opt_src = argv[i];
		} else if (!opt_dst) {
			opt_dst = argv[i];
		} else {
			error();
		}
	}

	if (!opt_src || !opt_dst)
		error();

	if (opt_scroll) {
		if (analyze_file(opt_src,&sc) != 0) {
			fprintf(stderr,"Error analyzing %s\n",opt_src);
			exit(EXIT_FAILURE);
		}

		if (convert_file(opt_src,opt_dst,&sc) != 0) {
			fprintf(stderr,"Error converting %s in %s\n",opt_src,opt_dst);
			exit(EXIT_FAILURE);
		}

	} else {

		if (convert_file(opt_src,opt_dst,0) != 0) {
			fprintf(stderr,"Error converting %s in %s\n",opt_src,opt_dst);
			exit(EXIT_FAILURE);
		}
	}

	return EXIT_SUCCESS;
}
